//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------

using UnityEngine;




namespace OQJ_KeyframeData 
{
	public class KeyframeData
	{
		int m_type;
		float m_frameNr;
		Vector3 m_position = Vector3.zero;
		Quaternion m_rotation = Quaternion.identity;

		Vector3 m_center = Vector3.zero;
		Vector3 m_axis = Vector3.forward;
		Vector3 m_startPosC = Vector3.zero;
		float m_degreeC;
		bool m_AN_CIR;
		float m_deg2;

		Vector3 m_axisStart = Vector3.zero;
		Vector3 m_centerAxis = Vector3.forward;
		Vector3 m_startPosH = Vector3.zero;
		float m_degreeH;
		float m_primaryLength;

		public static readonly int LINE = 21;
		public static readonly int CIRCLE = 22;
		public static readonly int SPIRAL = 23;

		static int pr_type;
		static Vector3 pr_center;
		static Vector3 pr_axis;
		static Vector3 pr_startPosC;
		static Vector3 pr_centerAxis;
		static Vector3 pr_axisStart;
		static Vector3 pr_startPosH;
		static float pr_primaryLength;

		static Vector3 pr_preKFPos;
		static float pr_preKFDeg;



		public KeyframeData(int type, Vector3 pos, Quaternion rot)
		{
			m_type = LINE;

			m_frameNr = 0;
			m_position = pos;
			m_rotation = rot;

			pr_type = m_type;

			pr_preKFPos = pos;
			pr_preKFDeg = 0.0f;
		}



		public KeyframeData(int type, Vector3 pos, Quaternion rot, Vector3 C, Vector3 axis) :
			this(0, type, pos, rot, C, axis) { }


		public KeyframeData(float nr, int type, Vector3 pos, Quaternion rot, Vector3 C, Vector3 axis)
		{
			if(pr_type == CIRCLE)
			{
				Vector3 u1 = (pr_preKFPos - pr_center).normalized;
				Vector3 u2 = (pos - pr_center).normalized;
				float cosT = Vector3.Dot(u1, u2);
				Vector3 cross = Vector3.Cross(u1, u2);  

				float deg = Mathf.Acos(cosT) * Mathf.Rad2Deg;  
				if (TH3DMath.IsParallel(axis, cross) == -1)
				{   
					deg = -deg;
				}

				m_deg2 = pr_preKFDeg + deg;		
				m_AN_CIR = true;
			}


			m_type = CIRCLE;

			m_frameNr = nr;
			m_position = pos;
			m_rotation = rot;
			m_degreeC = 0.0f;

			m_axis = axis;
			m_center = C;
			m_startPosC = m_position;

			pr_type = m_type;
			pr_center = m_center;
			pr_axis = m_axis;
			pr_startPosC = m_startPosC;


			pr_preKFPos = pos;
			pr_preKFDeg = m_degreeC;
		}



		public KeyframeData(int type, Vector3 pos, Quaternion rot, Vector3 C, Vector3 axis, float sn)
		{
			m_type = SPIRAL;

			m_frameNr = 0;
			m_position = pos;
			m_rotation = rot;

			m_centerAxis = axis;
			m_axisStart = C;
			m_primaryLength = sn;

			float d1 = Vector3.Dot(m_position - m_axisStart, m_centerAxis);

			if (Mathf.Abs(d1) > 0.001f)
			{
				Vector3 Q = m_position - d1 * m_centerAxis;  
				m_degreeH = -(d1 / m_primaryLength * 360.0f);

				m_startPosH = Quaternion.AngleAxis(-m_degreeH, axis) * (Q - m_axisStart) + m_axisStart;
			}
			else
			{
				m_startPosH = m_position;
				m_degreeH = 0.0f;
			}


			pr_type = m_type;
			pr_axisStart = m_axisStart;
			pr_centerAxis = m_centerAxis;
			pr_startPosH = m_startPosH;
			pr_primaryLength = m_primaryLength;
		}




		public KeyframeData(float nr, int type, Vector3 pos, Quaternion rot)
		{
			m_frameNr = nr;
			m_type = type;
			m_position = pos;
			m_rotation = rot;

			m_degreeC = 0.0f;
			m_degreeH = 0.0f;


			if (m_type == SPIRAL || pr_type == SPIRAL)
			{
				m_axisStart = pr_axisStart;
				m_centerAxis = pr_centerAxis;
				m_startPosH = pr_startPosH;
				m_primaryLength = pr_primaryLength;

				float d1 = Vector3.Dot(m_position - m_axisStart, m_centerAxis);
				m_degreeH = -(d1 / m_primaryLength * 360.0f);
			}
			else if (m_type == CIRCLE || pr_type == CIRCLE)		
			{
				m_center = pr_center;
				m_axis = pr_axis;
				m_startPosC = pr_startPosC;

				Vector3 u1 = (pr_preKFPos - m_center).normalized;
				Vector3 u2 = (m_position - m_center).normalized;
				float cosT = Vector3.Dot(u1, u2);
				Vector3 cross = Vector3.Cross(u1, u2);  

				float deg = Mathf.Acos(cosT) * Mathf.Rad2Deg;  
				if (TH3DMath.IsParallel(m_axis, cross) == -1)
				{   
					deg = -deg;
				}

				m_degreeC = pr_preKFDeg + deg;
			}
			else if (m_type == LINE)
			{
			}


			pr_type = m_type;
			pr_center = m_center;
			pr_axis = m_axis;

			pr_axisStart = m_axisStart;
			pr_centerAxis = m_centerAxis;
			pr_primaryLength = m_primaryLength;

			pr_preKFPos = pos;
			pr_preKFDeg = m_degreeC;
		}



		public int type
		{
			get { return m_type; }
		}


		public float frameNumber
		{
			get { return m_frameNr; }
		}


		public Vector3 position
		{
			get { return m_position; }
		}


		public Quaternion rotation
		{
			get { return m_rotation; }
		}


		public Vector3 center
		{
			get { return m_center; }
		}


		public Vector3 axis
		{
			get { return m_axis; }
		}


		public Vector3 startPosCircle
		{
			get { return m_startPosC; }
		}


		public Vector3 startPosHelix
		{
			get { return m_startPosH; }
		}


		public float degCircle
		{
			get { return m_degreeC; }
		}


		public bool anotherCircle
		{
			get { return m_AN_CIR; }
		}


		public float deg2
		{
			get { return m_deg2; }
		}



		public Vector3 centerAxis
		{
			get { return m_centerAxis; }
		}


		public Vector3 axisStart
		{
			get { return m_axisStart; }
		}


		public float degHelix
		{
			get { return m_degreeH; }
		}


		public float primaryLength
		{
			get { return m_primaryLength; }
		}
	}

} 

