//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------

using UnityEngine;


public static class TH2DMath
{
	static float ABOUT_ZERO = 0.00001f;

	
	public static Vector3 ToVector3(Vector2 v)
	{
		return new Vector3(v.x, v.y, 1.0f);
	}


	public static THMatrix3x3 GetTranslation3x3(float x, float y)	
	{ 
		return GetTranslation3x3(new Vector2(x, y));	
	}	


	public static THMatrix3x3 GetTranslation3x3(Vector2 tv)	
	{	
		Vector3 c0 = new Vector3(1.0f, 0.0f, 0.0f);	
		Vector3 c1 = new Vector3(0.0f, 1.0f, 0.0f);	
		Vector3 c2 = new Vector3(tv.x, tv.y, 1.0f);	
		THMatrix3x3 tra = new THMatrix3x3(c0, c1, c2);

		return tra;	
	}



	public static Quaternion GetRotation(float degree)
	{
		float rd2 = degree * 0.5f * Mathf.Deg2Rad;
		float s2 = Mathf.Sin(rd2);
		float c2 = Mathf.Cos(rd2);

		return new Quaternion(0, 0, s2, c2);
	}



	public static THMatrix3x3 GetRotation3x3(float degree)	
	{
		float angle = Mathf.Deg2Rad * degree;
		float cosT = Mathf.Cos(angle);
		float sinT = Mathf.Sin(angle);

		Vector3 c0 = new Vector3(cosT, sinT, 0.0f);	
		Vector3 c1 = new Vector3(-sinT, cosT, 0.0f);	
		Vector3 c2 = new Vector3(0.0f, 0.0f, 1.0f);	
		THMatrix3x3 rot = new THMatrix3x3(c0, c1, c2);

		return rot;	
	}


	public static THMatrix3x3 GetScale3x3(float s)	
	{
		return GetScale3x3(s, s) ;	
	}	


	public static THMatrix3x3 GetScale3x3(float x, float y)	
	{ 
		Vector3 c0 = new Vector3(x, 0.0f, 0.0f);	
		Vector3 c1 = new Vector3(0.0f, y, 0.0f);	
		Vector3 c2 = new Vector3(0.0f, 0.0f, 1.0f);	
		THMatrix3x3 scale = new THMatrix3x3(c0, c1, c2);

		return scale;	
	}	



	public static float CalcCameraPositionZ_2D(float worldL, float pixL=0.0f, bool pix_fit=false)
	{
		if (!pix_fit)
		{	
			float fov = BSData.CameraController2D.GetFieldOfView();
			float hhgt = worldL;
			float hfov = Mathf.Deg2Rad * (fov / 2.0f);      
			float vtan = Mathf.Tan(hfov);
			float dist = hhgt / vtan;
			return -dist;
		}
		else
		{	
			float fov = BSData.CameraController2D.GetFieldOfView();

			float upp = worldL / pixL;    
			float hhgt = BSData.ScreenH / 2.0f * upp;       
			float hfov = Mathf.Deg2Rad * (fov / 2.0f);      
			float vtan = Mathf.Tan(hfov);
			float dist = hhgt / vtan;

			return -dist;    
		}
	}



	public static int IsParallel(Vector2 v1, Vector2 v2)
	{
		Vector2 u1 = v1.normalized;
		Vector2 u2 = v2.normalized;

		float dotV = Vector2.Dot(u1, u2);

		if(THUtil.IsZero(1.0f - Mathf.Abs(dotV)))
		{
			if(dotV > ABOUT_ZERO)
			{
				return 1;	
			}
			else
			{
				return -1;
			}
		}

		return 0;	
	}



	public static bool IsParallel_BOOL(Vector2 v1, Vector2 v2)
	{
		Vector2 u1 = v1.normalized;
		Vector2 u2 = v2.normalized;

		float dotV = Mathf.Abs(Vector3.Dot(u1, u2));  

		if(THUtil.IsZero(1.0f - dotV))
		{
			return true;	
		}

		return false;	
	}



	public static Vector3 CalcIntersectionOf2Lines(Vector2 P, Vector2 a, Vector2 Q, Vector2 b)
	{
		a = a.normalized;
		b = b.normalized;

		Vector2 W = Q - P;
		float detA = -a.x * b.y + a.y * b.x;	
		if (!THUtil.IsZero(detA))
		{
			float s = (-b.y * W.x + b.x * W.y) / detA;
			Vector3 R = P + s * a;
			R.z = s;
			return R;
		}

		
		if(IsParallel_BOOL(W, a))
		{
			Vector3 R = (P + Q) * 0.5f;
			R.z = W.magnitude * 0.5f;
			return R;
		}

		return Vector3.positiveInfinity;
	}



	public static float[] CalcInverseMatrix2D(float a, float c, float b, float d)  
	{

		float det = a * d - b * c;
		if (THUtil.IsZero(det)) { return null; }

		float[] inv = { d / det, -c / det, -b / det, a / det };  

		return inv;
	}



	public static Quaternion GetRotationPart(THMatrix3x3 mtx)     
	{
		Vector2 c0 = mtx.GetColumn(0).normalized;
		Vector2 c1 = mtx.GetColumn(1).normalized;

		float det = c0.x * c1.y - c1.x * c0.y;	
		if(!THUtil.IsEqual(det, 1.0f)) 
		{
			Debug.Log("**** **** Conversion Error.. [TH2DMath_GetRotationPart(..)] **** ****");
			return Quaternion.identity; 
		}

		float cosT = c0.x;
		float sinT = c0.y;

		if(Mathf.Abs(sinT) < 0.001f)  
		{
			if(cosT > ABOUT_ZERO)	
			{
				return Quaternion.identity;
			}
			else					
			{
				return new Quaternion(0.0f, 0.0f, 1.0f, 0.0f);
			}
		}

		float s2 = Mathf.Sqrt((1.0f - cosT) * 0.5f);
		if(sinT < 0.0f) { s2 = -s2; }

		float c2 = (sinT / s2) * 0.5f;

		return new Quaternion(0.0f, 0.0f, s2, c2);
	}



	public static float CalcDegree_V1toV2(Vector2 v1, Vector2 v2)  
	{
		Vector2 u1 = v1.normalized;
		Vector2 u2 = v2.normalized;
		float dot = Vector2.Dot(u1, u2);
		float degree = (Mathf.Abs(dot) <= 0.999999f) ? (Mathf.Acos(dot) * Mathf.Rad2Deg) :
						 (dot > 0.0f ? 0.0f : 180.0f);     

		int sign = (u1.x * u2.y - u1.y * u2.x > 0.0f) ? 1 : -1;
		return sign * degree;
	}




	public static THMatrix3x3 CalcRotation3x3_V1toV2(Vector2 v1, Vector2 v2)
	{
		Vector2 u1 = v1.normalized;
		Vector2 u2 = v2.normalized;
		float cosT = Vector2.Dot(u1, u2);
		float sinT = Mathf.Sqrt(1.0f - cosT * cosT);

		sinT = (u1.x * u2.y - u1.y * u2.x > 0.0f) ? sinT : -sinT;

		Vector3 c0 = new Vector3(cosT, sinT, 0.0f);	
		Vector3 c1 = new Vector3(-sinT, cosT, 0.0f);	
		Vector3 c2 = new Vector3(0.0f, 0.0f, 1.0f);	
		THMatrix3x3 rot = new THMatrix3x3(c0, c1, c2);

		return rot;	
	}



	public static Quaternion CalcRotation_V1toV2(Vector2 V1, Vector2 V2)
	{
		float magV12 = V1.magnitude * V2.magnitude;  
		float vDot = Vector2.Dot(V1, V2);
		if(Mathf.Abs(magV12 - vDot) < 0.000001f)  
		{
			return Quaternion.identity;
		}
		else if(Mathf.Abs(magV12 + vDot) < 0.000001f)  
		{
			return new Quaternion(0, 0, 1, 0);		
		}

		float crsZ = V1.x * V2.y - V1.y * V2.x;
		Quaternion q = new Quaternion(0, 0, crsZ, vDot + magV12);  

		return q.normalized;
	}

} 

