//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------

using UnityEngine;


public static class TH3DMath
{
	static float ABOUT_ZERO = 0.00001f;

	
	public static Vector4 ToVector4(Vector3 v)
	{
		return new Vector4(v.x, v.y, v.z, 1.0f);
	}


	public static Vector3 ToVector3(Vector4 v)
	{
		return new Vector3(v.x, v.y, v.z);
	}


	public static Matrix4x4 GetTranslation4x4(Vector3 v)
	{
		return Matrix4x4.Translate(v);
	}


	public static Matrix4x4 GetTranslation4x4(float x, float y, float z)
	{
		return GetTranslation4x4(new Vector3(x, y, z));
	}


	public static Matrix4x4 GetRotation4x4(float deg, Vector3 axis)
	{

		Vector3 n = axis.normalized;
		float nx = n.x;
		float ny = n.y;
		float nz = n.z;
		float theta = Mathf.Deg2Rad * deg;
		float cosT = Mathf.Cos(theta);
		float sinT = Mathf.Sin(theta);

		Vector4 c1 = new Vector4( 
			nx * nx * (1 - cosT) + cosT, nx * ny * (1 - cosT) + nz * sinT, nx * nz * (1 - cosT) - ny * sinT, 0.0f
		);
		Vector4 c2 = new Vector4( 
			nx * ny * (1 - cosT) - nz * sinT, ny * ny * (1 - cosT) + cosT, ny * nz * (1 - cosT) + nx * sinT, 0.0f
		);
		Vector4 c3 = new Vector4( 
			nx * nz * (1 - cosT) + ny * sinT, ny * nz * (1 - cosT) - nx * sinT, nz * nz * (1 - cosT) + cosT, 0.0f
		);
		Vector4 c4 = new Vector4(0.0f, 0.0f, 0.0f, 1.0f); 

		return new Matrix4x4(c1, c2, c3, c4);	
	}	


	public static Matrix4x4 GetScale4x4(float s)	
	{
		return GetScale4x4(s, s, s); 
	}	


	public static Matrix4x4 GetScale4x4(float sx, float sy, float sz)	
	{ 
		Vector4 c1 = new Vector4(sx, 0.0f, 0.0f, 0.0f);	
		Vector4 c2 = new Vector4(0.0f, sy, 0.0f, 0.0f);	
		Vector4 c3 = new Vector4(0.0f, 0.0f, sz, 0.0f);	
		Vector4 c4 = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);	
		Matrix4x4 scale = new Matrix4x4(c1, c2, c3, c4);

		return scale;	
	}	


	public static Matrix4x4 GetEulerRotation4x4(float eulX, float eulY, float eulZ)
	{
		Quaternion rq = Quaternion.Euler(eulX, eulY, eulZ);
		return Matrix4x4.Rotate(rq);
	}	



	public static Quaternion GetRotationPart(Matrix4x4 mtx)		
	{
		Vector3 c0 = new Vector3(mtx[0, 0], mtx[1, 0], mtx[2, 0]).normalized;
		Vector3 c1 = new Vector3(mtx[0, 1], mtx[1, 1], mtx[2, 1]).normalized;
		Vector3 c2 = new Vector3(mtx[0, 2], mtx[1, 2], mtx[2, 2]).normalized;

		float m00 = c0.x, m10 = c0.y, m20 = c0.z;
		float m01 = c1.x, m11 = c1.y, m21 = c1.z;
		float m02 = c2.x, m12 = c2.y, m22 = c2.z;

		Quaternion q = Quaternion.identity;	

		float trace = m00 + m11 + m22 + 1.0f;	
		if (trace > ABOUT_ZERO)  
		{
			float s = 0.5f / Mathf.Sqrt(trace);     
			q.w = 0.25f / s;
			q.x = (m21 - m12) * s;
			q.y = (m02 - m20) * s;
			q.z = (m10 - m01) * s;
		}
		else    
		{
			float t;
			if ((t = m00 - m11 - m22 + 1.0f) > ABOUT_ZERO)
			{
				float s = 2.0f * Mathf.Sqrt(t);         
				q.w = 0.0f; 
				q.x = 0.25f * s;
				q.y = (m01 + m10) / s;
				q.z = (m02 + m20) / s;
			}
			else if ((t = -m00 + m11 - m22 + 1.0f) > ABOUT_ZERO)
			{
				float s = 2.0f * Mathf.Sqrt(t);         
				q.w = 0.0f; 
				q.x = (m01 + m10) / s;
				q.y = 0.25f * s;
				q.z = (m12 + m21) / s;
			}
			else if ((t = -m00 - m11 + m22 + 1.0f) > ABOUT_ZERO)
			{
				float s = 2.0f * Mathf.Sqrt(t);         
				q.w = 0.0f; 
				q.x = (m02 + m20) / s;
				q.y = (m12 + m21) / s;
				q.z = 0.25f * s;
			}
		}

		return q.normalized;
	}




	public static Vector3 CalcNormal(Vector3 A, Vector3 B, Vector3 C)
	{
		Vector3 VN = Vector3.Cross((B - A), (C - A));
		return VN.normalized;
	}


	public static Vector3 CalcNormal(Vector3[] vts)
	{
		if(vts == null || !(vts.Length == 3 || vts.Length == 4)) { return Vector3.zero; }

		return CalcNormal(vts[0], vts[1], vts[2]);
	}


	public static Quaternion CalcRotation_V1toV2(Vector3 V1, Vector3 V2)
	{
		return CalcRotation_V1toV2(V1, V2, Vector3.up);
	}


	public static Quaternion CalcRotation_V1toV2(Vector3 V1, Vector3 V2, Vector3 def_nrm)
	{
		float magV12 = V1.magnitude * V2.magnitude;  
		float vDot = Vector3.Dot(V1, V2);
		if(Mathf.Abs(magV12 - vDot) < 0.000001f)  
		{
			return Quaternion.identity;
		}
		else if(Mathf.Abs(magV12 + vDot) < 0.000001f)  
		{
			Vector3 N = def_nrm.normalized;
			return new Quaternion(N.x, N.y, N.z, 0.0f);		
		}

		Vector3 vCross = Vector3.Cross(V1, V2);  
		Quaternion q = new Quaternion(vCross.x, vCross.y, vCross.z, vDot + magV12);  

		return q.normalized;
	}



	public static float CalcDegree_V1toV2(Vector3 V1, Vector3 V2, Vector3 axis)  
	{
		Vector3 u1 = V1.normalized;
		Vector3 u2 = V2.normalized;
		float dot = Vector3.Dot(u1, u2);
		float degree = (Mathf.Abs(dot) <= 0.999999f) ? (Mathf.Acos(dot) * Mathf.Rad2Deg) :
						 (dot > 0.0f ? 0.0f : 180.0f);     

		Vector3 crs = Vector3.Cross(u1, u2);
		int sign = (Vector3.Dot(crs, axis) > 0.0f) ? 1 : -1;
		return sign * degree;
	}



	public static float CalcDegree_V1toV2(Vector3 V1, Vector3 V2)  
	{
		Vector3 u1 = V1.normalized;
		Vector3 u2 = V2.normalized;
		float dot = Vector3.Dot(u1, u2);
		float degree = (Mathf.Abs(dot) <= 0.999999f) ? (Mathf.Acos(dot) * Mathf.Rad2Deg) :
						 (dot > 0.0f ? 0.0f : 180.0f);     

		int sign = (u1.z * u2.x - u1.x * u2.z > 0.0f) ? 1 : -1;
		return sign * degree;
	}




	public static Quaternion CalcRotation_VW1toVW2(Vector3 v1, Vector3 w1, Vector3 v2, Vector3 w2, 
																			Vector3 def_nrm=new Vector3())
	{
		Vector3 uV1 = Vector3.Normalize(v1);
		Vector3 uW1 = Vector3.Normalize(w1);
		Vector3 uV2 = Vector3.Normalize(v2);
		Vector3 uW2 = Vector3.Normalize(w2);

		if (IsParallel_Unit(uV1, uV2) == 1 && IsParallel_Unit(uW1, uW2) == 1)
		{
			return Quaternion.identity;	
		}

		Quaternion wq = CalcRotation_V1toV2(uV1, uV2, def_nrm);
		Vector3 uW1R = wq * uW1;

		int parallel = IsParallel_Unit(uW1R, uW2);
		if(parallel == 1)
		{
			return wq;
		}
		else if(parallel == -1)
		{
			Quaternion rot180 = new Quaternion(uV2.x, uV2.y, uV2.z, 0.0f); 
			return rot180 * wq;
		}


		float dot = Vector3.Dot(uW1R, uW2);
		Vector3 crs = Vector3.Cross(uW1R, uW2);
		int sign = (Vector3.Dot(crs, uV2) > 0) ? 1 : -1;    
		Vector3 axis = sign * uV2;

		float s2 = Mathf.Sqrt((1.0f - dot) * 0.5f);		
		float c2 = crs.magnitude / (2 * s2);
		Quaternion wq2 = new Quaternion(s2 * axis.x, s2 * axis.y, s2 * axis.z, c2);

		return (wq2 * wq).normalized;
	}



	public static Quaternion CalcRotation_VW1toVW2_HighPrec(Vector3 v1, Vector3 w1, Vector3 v2, Vector3 w2, 
																			Vector3 def_nrm=new Vector3())
	{
		Vector3 uV1 = Vector3.Normalize(v1);
		Vector3 uW1 = Vector3.Normalize(w1);
		Vector3 uV2 = Vector3.Normalize(v2);
		Vector3 uW2 = Vector3.Normalize(w2);

		if (IsParallel_Unit_HighPrec(uV1, uV2) == 1 && IsParallel_Unit_HighPrec(uW1, uW2) == 1)
		{
			return Quaternion.identity;	
		}

		Quaternion wq = CalcRotation_V1toV2(uV1, uV2, def_nrm);
		Vector3 uW1R = wq * uW1;

		int parallel = IsParallel_Unit_HighPrec(uW1R, uW2);
		if(parallel == 1)
		{
			return wq;
		}
		else if(parallel == -1)
		{
			Quaternion rot180 = new Quaternion(uV2.x, uV2.y, uV2.z, 0.0f); 
			return rot180 * wq;
		}


		float dot = Vector3.Dot(uW1R, uW2);
		Vector3 crs = Vector3.Cross(uW1R, uW2);
		int sign = (Vector3.Dot(crs, uV2) > 0) ? 1 : -1;    
		Vector3 axis = sign * uV2;

		float s2 = Mathf.Sqrt((1.0f - dot) * 0.5f);		
		float c2 = crs.magnitude / (2 * s2);
		Quaternion wq2 = new Quaternion(s2 * axis.x, s2 * axis.y, s2 * axis.z, c2);

		return (wq2 * wq).normalized;
	}



	public static Matrix4x4 LookAtLH(Vector3 posi, Vector3 target, Vector3 up)
	{   
		Vector3 uK = Vector3.Normalize(target - posi);
		Vector3 uI = Vector3.Normalize(Vector3.Cross(up, uK));
		Vector3 uJ = Vector3.Cross(uK, uI);

		Matrix4x4 wrot = new Matrix4x4(new Vector4(uI.x, uJ.x, uK.x, 0.0f), 
										new Vector4(uI.y, uJ.y, uK.y, 0.0f),    
										new Vector4(uI.z, uJ.z, uK.z, 0.0f),    
										new Vector4(0.0f, 0.0f, 0.0f, 1.0f));   

		Matrix4x4 wtrl = GetTranslation4x4(-posi);

		return wrot * wtrl;
	}


	public static int IsParallel(Vector3 v1, Vector3 v2)
	{
		Vector3 u1 = Vector3.Normalize(v1);
		Vector3 u2 = Vector3.Normalize(v2);

		float dotV = Vector3.Dot(u1, u2);

		if(THUtil.IsZero(1.0f - Mathf.Abs(dotV)))
		{
			if(dotV > ABOUT_ZERO)
			{
				return 1;	
			}
			else
			{
				return -1;
			}
		}

		return 0;	
	}



	public static int IsParallel_Unit(Vector3 u1, Vector3 u2)
	{
		float dotV = Vector3.Dot(u1, u2);

		if(THUtil.IsZero(1.0f - Mathf.Abs(dotV)))
		{
			if(dotV > ABOUT_ZERO)
			{
				return 1;	
			}
			else
			{
				return -1;
			}
		}

		return 0;	
	}



	public static bool IsParallel_BOOL(Vector3 v1, Vector3 v2)
	{
		Vector3 u1 = Vector3.Normalize(v1);
		Vector3 u2 = Vector3.Normalize(v2);

		float dotV = Mathf.Abs(Vector3.Dot(u1, u2));  

		if(THUtil.IsZero(1.0f - dotV))
		{
			return true;	
		}

		return false;	
	}



	public static int IsParallel_Unit_HighPrec(Vector3 u1, Vector3 u2)
	{
		float dotV = Vector3.Dot(u1, u2);

		if(THUtil.IsZero_HighPrec(1.0f - Mathf.Abs(dotV)))
		{
			if(dotV > ABOUT_ZERO)
			{
				return 1;	
			}
			else
			{
				return -1;
			}
		}

		return 0;	
	}



	public static Vector3 ConvertPositionByBasis(Vector3 src_pos, Vector3 basis_orig, 
												Vector3 basis_x, Vector3 basis_y, Vector3 basis_z)
	{
		Vector3 w = src_pos - basis_orig;
		float a = Vector3.Dot(w, basis_x);
		float b = Vector3.Dot(w, basis_y);
		float c = Vector3.Dot(w, basis_z);
		return new Vector3(a, b, c);
	}


	public static Vector3 ConvertDirectionByBasis(Vector3 src_dir, Vector3 basis_x, Vector3 basis_y, Vector3 basis_z)
	{
		float a = Vector3.Dot(src_dir, basis_x);
		float b = Vector3.Dot(src_dir, basis_y);
		float c = Vector3.Dot(src_dir, basis_z);
		return new Vector3(a, b, c).normalized;
	}

} 

