//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------

using UnityEngine;



namespace OQJ_THCollisionTest
{

	public static class THCollisionTest
	{
		public static bool CollisionTest_Disk_Disk(Vector2 C1, float r1, Vector2 C2, float r2)
		{
			float l = (C1 - C2).magnitude;

			if (r1 + r2 >= l)
			{
				return true;
			}

			return false;
		}



		public static bool CollisionTest_Disk_Rect(Vector2 C, float r, Vector2 T, float w, float h,
													Vector2 lx, Vector2 ly)
		{
			float w2 = w * 0.5f;
			float h2 = h * 0.5f;

			Vector2 wv = C - T;
			float s = Mathf.Abs(Vector2.Dot(lx, wv));
			float t = Mathf.Abs(Vector2.Dot(ly, wv));
			Vector2 locC = new Vector2(s, t);

			if ((locC.x >= w2 + r) || (locC.y >= h2 + r))
			{
				return false;
			}

			float ax = Mathf.Clamp(locC.x - w2, 0.0f, r);
			float ay = Mathf.Clamp(locC.y - h2, 0.0f, r);
			if (ax * ax + ay * ay >= r * r)
			{
				return false;
			}

			return true;
		}



		public static bool CollisionTest_Rect_Rect(Vector2 T1, Vector2 lx1, Vector2 ly1, float w1, float h1,
													Vector2 T2, Vector2 lx2, Vector2 ly2, float w2, float h2)
		{
			float ww = w1 * 0.5f;
			float hh = h1 * 0.5f;
			Vector2[] verts1 =
			{
				T1 + ww * lx1 + hh * ly1,
				T1 + ww * lx1 - hh * ly1,
				T1 - ww * lx1 - hh * ly1,
				T1 - ww * lx1 + hh * ly1
			};

			ww = w2 * 0.5f;
			hh = h2 * 0.5f;
			Vector2[] verts2 =
			{
				T2 + ww * lx2 + hh * ly2,
				T2 + ww * lx2 - hh * ly2,
				T2 - ww * lx2 - hh * ly2,
				T2 - ww * lx2 + hh * ly2
			};

			return CollisionTest_Rect_Rect(verts1, verts2);
		}


		public static bool CollisionTest_Rect_Rect(Vector2[] vertsA, Vector2[] vertsB)
		{
			Vector2[] sideDirs =
			{
				vertsA[0] - vertsA[1], vertsA[1] - vertsA[2],
				vertsB[0] - vertsB[1], vertsB[1] - vertsB[2]
			};
			
			foreach (var u in sideDirs)
			{
				float minA = Vector2.Dot(vertsA[0], u);
				float maxA = minA;
				float minB = Vector2.Dot(vertsB[0], u);
				float maxB = minB;
				
				for (int i = 1; i < 4; i++)
				{
					float projA = Vector2.Dot(vertsA[i], u);
					if (projA < minA)		{ minA = projA; }
					else if (projA > maxA)	{ maxA = projA; }
					
					float projB = Vector2.Dot(vertsB[i], u);
					if (projB < minB)		{ minB = projB;	}
					else if (projB > maxB)	{ maxB = projB; }
				}
				
				if (maxB < minA || maxA < minB)
				{
					return false;
				}
			}
			
			return true;
		}


		public static Vector3 CollisionTest_Ray_Rect_Coplanar(Vector2 R0, Vector2 v,    // ** v : unit **
												Vector2 T, Vector2 ui, Vector2 uj, float w, float h)
		{
			Vector2 m = R0 - T;
			float s = Vector2.Dot(m, ui);
			float t = Vector2.Dot(m, uj);
			Vector2 locR0 = new Vector2(s, t);

			s = Vector2.Dot(v, ui);
			t = Vector2.Dot(v, uj);
			Vector2 locV = new Vector2(s, t);

			Vector2 max = new Vector2(w / 2, h / 2);
			Vector2 min = -max;
			Vector3 locP = CollisionTest_Ray_LocalRect_Coplanar(locR0, locV, min, max);
			if (locP.z < 0.0f)
			{
				return new Vector3(0, 0, -1);
			}

			Vector3 P = T + locP.x * ui + locP.y * uj;
			P.z = locP.z;
			return P;
		}



		public static Vector3 CollisionTest_Ray_LocalRect_Coplanar(Vector2 R0, Vector2 v, Vector2 min, Vector2 max)  // v : unit
		{
			float xmin = min.x;
			float ymin = min.y;
			float xmax = max.x;
			float ymax = max.y;

			//Vector2 u = v.normalized;

			float tx0, tx1;
			if (THUtil.IsZero(v.x))
			{
				tx0 = float.NegativeInfinity;
				tx1 = float.PositiveInfinity;
				if (R0.x < xmin || xmax < R0.x) { return new Vector3(0, 0, -1); }
			}
			else
			{
				tx0 = (xmin - R0.x) / v.x;
				tx1 = (xmax - R0.x) / v.x;
				if (v.x < -0.000001f) { THUtil.Swap(ref tx0, ref tx1); }
			}

			float ty0, ty1;
			if (THUtil.IsZero(v.y))
			{
				ty0 = float.NegativeInfinity;
				ty1 = float.PositiveInfinity;
				if (R0.y < ymin || ymax < R0.y) { return new Vector3(0, 0, -1); }
			}
			else
			{
				ty0 = (ymin - R0.y) / v.y;
				ty1 = (ymax - R0.y) / v.y;
				if (v.y < -0.000001f) { THUtil.Swap(ref ty0, ref ty1); }
			}


			if (tx0 > ty1 || tx1 < ty0) { return new Vector3(0, 0, -1); }

			if (tx0 < 0.0f && ty0 < 0.0f) { return new Vector3(0, 0, -1); }

			float tmin = Mathf.Max(tx0, ty0);
			float tmax = Mathf.Min(tx1, ty1);

			Vector3 P = R0 + tmin * v;
			P.z = tmin;
			return P;
		}



		public static Vector3 CollisionTest_Ray_Disk_Coplanar(Vector2 R0, Vector2 v, Vector2 C, float r)  // v : unit
		{
			Vector2 locR0 = R0 - C;
			float b = Vector2.Dot(locR0, v);
			float c = locR0.sqrMagnitude - r * r;

			float D = b * b - c;
			if (D < 0.0f) { return new Vector3(0, 0, -1); }

			float tmin = -b - Mathf.Sqrt(b * b - c);
			float tmax = -b + Mathf.Sqrt(b * b - c);

			if (tmin < 0.0f) { return new Vector3(0, 0, -1); }

			Vector3 P = R0 + tmin * v;
			P.z = tmin;
			return P;
		}

	}

}

