﻿//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------


using UnityEngine;

public struct THMatrix3x3
{
	public readonly float m00;
	public readonly float m01;
	public readonly float m02;

	public readonly float m10;
	public readonly float m11;
	public readonly float m12;

	public readonly float m20;
	public readonly float m21;
	public readonly float m22;


	public readonly float determinant;


	readonly Vector3 invCol0;
	readonly Vector3 invCol1;
	readonly Vector3 invCol2;

	public THMatrix3x3 inverse {
		get
		{
			return new THMatrix3x3(invCol0, invCol1, invCol2);
		}	
	}
	

	public static readonly THMatrix3x3 identity = new THMatrix3x3(
		new Vector3(1.0f, 0.0f, 0.0f), 
		new Vector3(0.0f, 1.0f, 0.0f), 
		new Vector3(0.0f, 0.0f, 1.0f)
	);	


	public THMatrix3x3(Vector3 col0, Vector3 col1, Vector3 col2)
	{
		//// col 0	
		m00 = col0.x;	
		m10 = col0.y;	
		m20 = col0.z;	

		//// col 1	
		m01 = col1.x;	
		m11 = col1.y;	
		m21 = col1.z;	

		//// col 2	
		m02 = col2.x;	
		m12 = col2.y;	
		m22 = col2.z;


		determinant  = (m00 * m11 * m22) + (m01 * m12 * m20) + (m02 * m10 * m21) 
					- (m02 * m11 * m20) - (m01 * m10 * m22) - (m00 * m12 * m21);	


		float inv_det = 1.0f / determinant;
		invCol0 = (new Vector3(m11*m22 - m12*m21, -(m10*m22 - m12*m20), m10*m21 - m11*m20)) * inv_det;
		invCol1 = (new Vector3(-(m01 * m22 - m02 * m21), m00 * m22 - m02 * m20, -(m00 * m21 - m01 * m20))) * inv_det; ;
		invCol2 = (new Vector3(m01 * m12 - m02 * m11, -(m00 * m12 - m02 * m10), m00 * m11 - m01 * m10)) * inv_det;
	}

		
	public Vector3 GetRow(int index)
	{
		return new Vector3(this[index, 0], this[index, 1], this[index, 2]);	
	}		


	public Vector3 GetColumn(int index)
	{
		return new Vector3(this[0, index], this[1, index], this[2, index]);	
	}		



	public static THMatrix3x3 operator* (THMatrix3x3 mtx1, THMatrix3x3 mtx2)
	{
		// col 0	
		float p00 = Vector3.Dot(mtx1.GetRow(0), mtx2.GetColumn(0)); 
		float p10 = Vector3.Dot(mtx1.GetRow(1), mtx2.GetColumn(0)); 
		float p20 = Vector3.Dot(mtx1.GetRow(2), mtx2.GetColumn(0)); 

		// col 1	
		float p01 = Vector3.Dot(mtx1.GetRow(0), mtx2.GetColumn(1)); 
		float p11 = Vector3.Dot(mtx1.GetRow(1), mtx2.GetColumn(1)); 
		float p21 = Vector3.Dot(mtx1.GetRow(2), mtx2.GetColumn(1)); 

		// col 2	
		float p02 = Vector3.Dot(mtx1.GetRow(0), mtx2.GetColumn(2)); 
		float p12 = Vector3.Dot(mtx1.GetRow(1), mtx2.GetColumn(2)); 
		float p22 = Vector3.Dot(mtx1.GetRow(2), mtx2.GetColumn(2));

		THMatrix3x3 rsltM = new THMatrix3x3(new Vector3(p00, p10, p20),
											new Vector3(p01, p11, p21), 
											new Vector3(p02, p12, p22));
		return rsltM;	
	}	


	public static Vector3 operator* (THMatrix3x3 mtx, Vector3 vtr)
	{
		Vector3 rsltV = new Vector3();

		rsltV.x = Vector3.Dot(mtx.GetRow(0), vtr); 
		rsltV.y = Vector3.Dot(mtx.GetRow(1), vtr); 
		rsltV.z = Vector3.Dot(mtx.GetRow(2), vtr); 

		return rsltV;	
	}	



	public float this[int row, int col]
	{
		get
		{
			if(row == 0)
			{
				if		(col == 0) { return m00; }
				else if (col == 1) { return m01; }
				else if (col == 2) { return m02; }
			}
			else if(row == 1)
			{
				if		(col == 0) { return m10; }
				else if (col == 1) { return m11; }
				else if (col == 2) { return m12; }
			}
			else if(row == 2)
			{
				if		(col == 0) { return m20; }
				else if (col == 1) { return m21; }
				else if (col == 2) { return m22; }
			}

			return float.PositiveInfinity;	
		}
	}	



	public override string ToString()
	{
		string row0 = string.Format("{0:f4},\t{1:f4},\t{2:f4}", this[0, 0], this[0, 1], this[0, 2]);
		string row1 = string.Format("{0:f4},\t{1:f4},\t{2:f4}", this[1, 0], this[1, 1], this[1, 2]);
		string row2 = string.Format("{0:f4},\t{1:f4},\t{2:f4}", this[2, 0], this[2, 1], this[2, 2]);

		return row0 + "\n" + row1 + "\n" + row2 + "\n";	
	}

}
