//-------------------------------------------
// Copyright (c) 2020-2025, Redpoll's 60.
// https://www.eproph527tw.com/
//-------------------------------------------

using System.Collections.Generic;
using UnityEngine;


namespace OQJ_THTransform
{
	public class THTransform
	{
		private THObject3D m_object;

		private Vector3 m_localPosition = new Vector3(0.0f, 0.0f, 0.0f);
		private Matrix4x4 m_localRotation = Matrix4x4.identity;
		private Vector3 m_localScale = new Vector3(1.0f, 1.0f, 1.0f);

		private Matrix4x4 m_localMatrix = Matrix4x4.identity;
		private Matrix4x4 m_worldMatrix = Matrix4x4.identity;

		private bool m_UPDATE_LOCAL = false;
		private bool m_UPDATE_WORLD = false;
		private bool m_EXE_WORLD = false;

		private THTransform m_parent = null;
		private List<THTransform> m_children = new List<THTransform>();


		public THTransform(THObject3D o)
		{
			m_object = o;
		}

		public Vector3 thLocalPosition
		{
			get { return m_localPosition; }

			set
			{
				m_localPosition = value;

				m_UPDATE_LOCAL = true;
				NotifyChain();
			}
		}

		public Matrix4x4 thLocalRotation
		{
			get { return m_localRotation; }

			set
			{
				m_localRotation = value;

				m_UPDATE_LOCAL = true;
				NotifyChain();
			}
		}

		public Vector3 thLocalScale
		{
			get { return m_localScale; }

			set
			{
				m_localScale = value;

				m_UPDATE_LOCAL = true;
				NotifyChain();
			}
		}


		public Matrix4x4 GetLocalMatrix()
		{
			if (m_UPDATE_LOCAL)
			{
				m_localMatrix = CalcLocalMatrix();
				m_UPDATE_LOCAL = false;
			}

			return m_localMatrix;
		}


		public Matrix4x4 GetWorldMatrix()
		{
			if (m_UPDATE_WORLD)
			{
				m_worldMatrix = CalcWorldMatrix();
				m_UPDATE_WORLD = false;
			}

			return m_worldMatrix;
		}


		private void NotifyChain()
		{
			m_UPDATE_WORLD = true;
			m_EXE_WORLD = true;

			foreach (var child in m_children)
			{
				child.NotifyChain();
			}
		}


		public void UpdateChain()
		{
			UpdateObject();

			foreach (var child in m_children)
			{
				child.UpdateChain();
			}
		}


		private void UpdateObject()
		{
			if (m_EXE_WORLD)
			{
				Matrix4x4 worldM = GetWorldMatrix();
				m_object.SetMatrix(worldM);
				m_EXE_WORLD = false;
			}
		}


		private Matrix4x4 CalcLocalMatrix()
		{
			float tx = m_localPosition.x;
			float ty = m_localPosition.y;
			float tz = m_localPosition.z;
			float sx = m_localScale.x;
			float sy = m_localScale.y;
			float sz = m_localScale.z;
			Matrix4x4 r = m_localRotation;
			Vector4 c1 = new Vector4(sx * r[0, 0], sx * r[1, 0], sx * r[2, 0], 0.0f);
			Vector4 c2 = new Vector4(sy * r[0, 1], sy * r[1, 1], sy * r[2, 1], 0.0f);
			Vector4 c3 = new Vector4(sz * r[0, 2], sz * r[1, 2], sz * r[2, 2], 0.0f);
			Vector4 c4 = new Vector4(tx, ty, tz, 1.0f);

			return new Matrix4x4(c1, c2, c3, c4);
		}


		private Matrix4x4 CalcWorldMatrix()
		{
			Matrix4x4 localM = GetLocalMatrix();

			Matrix4x4 worldM;
			if (m_parent == null)
			{
				worldM = localM; 
			}
			else
			{
				worldM = m_parent.GetWorldMatrix() * localM;
			}

			return worldM;
		}



		public void SetParent(THTransform p)
		{
			if (p == null)
			{
				if (m_parent != null)
				{
					m_parent.m_children.Remove(this);
					m_parent = null;
				}

				return;
			}

			if (p == this)
			{
				Debug.Log("Error : 1 [THTransform_SetParent()]");
				return;
			}

			if (m_parent != null)
			{
				Debug.Log("Error : 2 [THTransform_SetParent()]");
				return;
			}

			THTransform w = p;
			while (true)
			{
				if (w.m_parent == null) { break; }

				if (w.m_parent == this)
				{
					Debug.Log("Error : 3 [THTransform_SetParent()]");
					return;
				}

				w = w.m_parent;
			}


			m_parent = p;
			m_parent.m_children.Add(this);
		}


		public bool HasParent()
		{
			return (m_parent != null);
		}



		public void Clear()  
		{
			m_parent = null;
			m_children = new List<THTransform>();

			m_localPosition = new Vector3(0.0f, 0.0f, 0.0f);
			m_localRotation = Matrix4x4.identity;
			m_localScale = new Vector3(1.0f, 1.0f, 1.0f);

			m_localMatrix = Matrix4x4.identity;
			m_worldMatrix = Matrix4x4.identity;

			m_UPDATE_LOCAL = false;
			m_UPDATE_WORLD = false;
			m_EXE_WORLD = false;
		}

	}	

} // Elsa

